predicate partitioned(a: array<int>, k: nat, l: nat, m: nat, n: nat) 
reads a
{
    forall i: nat, j: nat :: k <= i < l <= m <= j < n <= a.Length ==> a[i] <= a[j]
}

predicate sorted(a: array<int>, l: nat, u: nat)
reads a
requires 0 <= l <= u <= a.Length
{
  forall j, k :: l <= j <= k < u ==> a[j] <= a[k]
}

predicate eq(a: array<int>, b: array<int>, l: nat, u: nat) 
reads a
reads b
requires 0 <= l <= u <= a.Length
requires a.Length == b.Length
{
    forall i: nat :: l <= i < u ==> a[i] == b[i]
}

method quicksort(a: array<int>)
ensures sorted(a, 0, a.Length)

modifies a
//ensures a.Length == r.Length
{
    if (a.Length > 1) {
    	qsort(a, 0, a.Length);
    }
}

method qsort(a: array<int>, l: nat, u: nat)
requires a.Length > 1
requires 0 <= l <= u <= a.Length
requires partitioned(a, 0, l, l, u)
requires partitioned(a, l, u, u, a.Length)

ensures sorted(a, l, u)
ensures forall i: nat :: 0 <= i < l ==> a[i] == old(a[i])
ensures forall i: nat :: u <= i < a.Length ==> a[i] == old(a[i])
ensures partitioned(a, 0, l, l, u)
ensures partitioned(a, l, u, u, a.Length)

decreases u - l
modifies a
{
    if l >= u - 1 {
    } else {
        var p_pivot := partition(a, l, u);
        qsort(a, l, p_pivot);
        qsort(a, p_pivot + 1, u);
    }
}

method partition(a: array<int>, l: nat, u: nat) returns(r_pivot: nat)
requires 0 <= l < u <= a.Length
requires partitioned(a, 0, l, l, u)
requires partitioned(a, l, u, u, a.Length)

ensures partitioned(a, l, r_pivot, r_pivot, u)
ensures partitioned(a, 0, l, l, u)
ensures partitioned(a, l, u, u, a.Length)
ensures partitioned(a, l, r_pivot, r_pivot, r_pivot + 1)
ensures partitioned(a, r_pivot, r_pivot + 1, r_pivot + 1, u)

ensures forall i: nat :: 0 <= i < l ==> a[i] == old(a[i])
ensures forall i: nat :: u <= i < a.Length ==> a[i] == old(a[i])
ensures l <= r_pivot < u

modifies a
{
    var pivot: nat := random(l, u);
    var pv : int := a[pivot];
    a[pivot] := a[u - 1];
    a[u - 1] := pv;
    
    var i: int := l - 1;
    var j: int := l;
    while j < u - 1 
        decreases u - j
        invariant i <= j < u
        invariant i < u - 1
        
        invariant forall x :: ((x>=l && x<=i) ==> a[x]<=pv)
        invariant forall x :: ((x>i && x<j) ==> a[x]>=pv)
        invariant j>i && i>=l-1 && j<u && a[u - 1]==pv && i>=-1 && l<u && l>=0
        invariant partitioned(a, 0, l, l, u)
        invariant partitioned(a, l, u, u, a.Length)
        invariant forall i: nat :: 0 <= i < l ==> a[i] == old(a[i])
        invariant forall i: nat :: u <= i < a.Length ==> a[i] == old(a[i])
    
    {
        if a[j] <= pv {
            i := i + 1;
            var t: int := a[i];
            a[i] := a[j];
            a[j] := t;
        }
        
        j := j + 1;
    }
    
    var t: int := a[i + 1];
    a[i + 1] := pv;
    a[u - 1] := t;
    
    return i + 1;

}

method random(l: nat, u: nat) returns (r: nat) 
requires l < u

ensures l <= r < u
{
    return (l + u) / 2;
}

method print_array(data: array<int>) 
{
    var i: nat := 0;
    
    while(i < data.Length) {
        print(data[i]);
        i := i + 1;
    }
    print("\n");
}

method Main() {
    var data: array<int> := new int[9](i requires 0 <= i < 9 => [0, 3, 5, 2, 1, 6, 4, 8, 7][i]);
    
    print_array(data);
    quicksort(data);
    print_array(data);
}

